"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NgJestCompilerHost = void 0;
const path_1 = require("path");
const constants_1 = require("ts-jest/dist/constants");
class NgJestCompilerHost {
    constructor(logger, ngJestCfg, jestCacheFS) {
        this.logger = logger;
        this.ngJestCfg = ngJestCfg;
        this.jestCacheFS = jestCacheFS;
        this._sourceFileCache = new Map();
        this._emittedResult = ['', ''];
        this._ts = this.ngJestCfg.compilerModule;
        this._moduleResolutionCache = this._ts.createModuleResolutionCache(this.ngJestCfg.cwd, (x) => x);
    }
    getEmittedResult() {
        return this._emittedResult;
    }
    updateMemoryHost(fileName, fileContent) {
        const previousContents = this.jestCacheFS.get(fileName);
        const contentsChanged = previousContents !== fileContent;
        if (contentsChanged) {
            this.jestCacheFS.set(fileName, fileContent);
        }
    }
    getSourceFile(fileName, languageVersion) {
        const normalizedFileName = path_1.normalize(fileName);
        try {
            const cached = this._sourceFileCache.get(normalizedFileName);
            if (cached) {
                return cached;
            }
            this.logger.debug({ normalizedFileName }, 'getSourceFile: cache miss, reading file and create source file to update cache');
            const content = this.readFile(normalizedFileName);
            if (content !== undefined) {
                const sf = this._ts.createSourceFile(normalizedFileName, content, languageVersion, true);
                this._sourceFileCache.set(fileName, sf);
                return sf;
            }
        }
        catch (e) {
            this.logger.error(e);
        }
        return undefined;
    }
    getDefaultLibFileName(options) {
        return this._ts.createCompilerHost(options).getDefaultLibFileName(options);
    }
    get writeFile() {
        return (fileName, data, _writeByteOrderMark) => {
            this._emittedResult[fileName.endsWith('.map') ? 1 : 0] = data;
        };
    }
    getCurrentDirectory() {
        return this.ngJestCfg.cwd;
    }
    getCanonicalFileName(fileName) {
        return this.useCaseSensitiveFileNames() ? fileName : fileName.toLowerCase();
    }
    useCaseSensitiveFileNames() {
        return this._ts.sys.useCaseSensitiveFileNames;
    }
    getNewLine() {
        return constants_1.LINE_FEED;
    }
    fileExists(fileName) {
        return this._ts.sys.fileExists(fileName);
    }
    readFile(fileName) {
        var _a;
        const normalizedFileName = path_1.normalize(fileName);
        let fileContent = this.jestCacheFS.get(normalizedFileName);
        this.logger.debug({ fileName: normalizedFileName }, 'readFile: file does not exist in memory cache, read file with file system');
        if (!fileContent) {
            fileContent = (_a = this._ts.sys.readFile(normalizedFileName)) !== null && _a !== void 0 ? _a : undefined;
            if (fileContent) {
                this.jestCacheFS.set(normalizedFileName, fileContent);
            }
        }
        return fileContent;
    }
    trace(message) {
        this.logger.trace(message);
    }
    resolveModuleNames(moduleNames, containingFile, _reusedNames, redirectedReference, options) {
        return moduleNames.map((moduleName) => {
            const { resolvedModule } = this._ts.resolveModuleName(moduleName, containingFile, options, this._ts.createCompilerHost(options), this._moduleResolutionCache, redirectedReference);
            return resolvedModule;
        });
    }
    resolveTypeReferenceDirectives(typeReferenceDirectiveNames, containingFile, redirectedReference, options) {
        return typeReferenceDirectiveNames.map((typeDirectiveName) => {
            const { resolvedTypeReferenceDirective } = this._ts.resolveTypeReferenceDirective(typeDirectiveName, containingFile, options, this._ts.createCompilerHost(options), redirectedReference);
            return resolvedTypeReferenceDirective;
        });
    }
}
exports.NgJestCompilerHost = NgJestCompilerHost;
