"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NgJestCompiler = void 0;
const compiler_host_1 = require("@angular/compiler-cli/src/transformers/compiler_host");
const compiler_utils_1 = require("ts-jest/dist/compiler/compiler-utils");
const ts_compiler_1 = require("ts-jest/dist/compiler/ts-compiler");
const compiler_host_2 = require("./compiler-host");
class NgJestCompiler extends ts_compiler_1.TsCompiler {
    constructor(ngJestConfig, jestCacheFS) {
        super(ngJestConfig, jestCacheFS);
        this.ngJestConfig = ngJestConfig;
        this.jestCacheFS = jestCacheFS;
        this._rootNames = [];
        this._ts = this.ngJestConfig.compilerModule;
        this._setupOptions(this.ngJestConfig);
    }
    getResolvedModulesMap(fileContent, fileName) {
        var _a, _b, _c;
        (_a = this._tsHost) === null || _a === void 0 ? void 0 : _a.updateMemoryHost(fileName, fileContent);
        return (_c = (_b = this._program) === null || _b === void 0 ? void 0 : _b.getSourceFile(fileName)) === null || _c === void 0 ? void 0 : _c.resolvedModules;
    }
    getCompiledOutput(fileContent, fileName, supportsStaticESM) {
        const customTransformers = this.ngJestConfig.resolvedTransformers;
        let moduleKind = this._initialCompilerOptions.module;
        let esModuleInterop = this._initialCompilerOptions.esModuleInterop;
        let allowSyntheticDefaultImports = this._initialCompilerOptions.allowSyntheticDefaultImports;
        if (supportsStaticESM && this.ngJestConfig.useESM) {
            moduleKind =
                !moduleKind ||
                    (moduleKind &&
                        ![this._ts.ModuleKind.ES2015, this._ts.ModuleKind.ES2020, this._ts.ModuleKind.ESNext].includes(moduleKind))
                    ? this._ts.ModuleKind.ESNext
                    : moduleKind;
            esModuleInterop = true;
            allowSyntheticDefaultImports = true;
        }
        else {
            moduleKind = this._ts.ModuleKind.CommonJS;
        }
        this._compilerOptions = Object.assign(Object.assign({}, this._compilerOptions), { allowSyntheticDefaultImports,
            esModuleInterop, module: moduleKind });
        if (!this.ngJestConfig.isolatedModules) {
            return super.getCompiledOutput(fileContent, fileName, supportsStaticESM);
        }
        else {
            this._logger.debug({ fileName }, 'getCompiledOutput: compiling as isolated module');
            const result = this._transpileModule(fileContent, {
                fileName,
                compilerOptions: this._compilerOptions,
            }, customTransformers);
            if (result.diagnostics && this.ngJestConfig.shouldReportDiagnostics(fileName)) {
                this.ngJestConfig.raiseDiagnostics(result.diagnostics, fileName, this._logger);
            }
            return compiler_utils_1.updateOutput(result.outputText, fileName, result.sourceMapText);
        }
    }
    _setupOptions({ parsedTsConfig }) {
        this._logger.debug({ parsedTsConfig }, '_setupOptions: initializing compiler config');
        this._compilerOptions = Object.assign({}, this._initialCompilerOptions);
        this._rootNames = parsedTsConfig.rootNames.filter((rootName) => !this.ngJestConfig.isTestFile(rootName));
        if (this._compilerOptions.strictMetadataEmit) {
            this._logger.warn(`Using Angular compiler option 'strictMetadataEmit' for applications might cause undefined behavior.`);
        }
        if (!this.ngJestConfig.isolatedModules) {
            this._logger.debug({ compilerOptions: this._compilerOptions }, '_setupOptions: creating Compiler Host using @angular/compiler-cli createCompilerHost');
            this._tsHost = new compiler_host_2.NgJestCompilerHost(this._logger, this.ngJestConfig, this.jestCacheFS);
        }
    }
    _createOrUpdateProgram() {
        const oldTsProgram = this._program;
        this._logger.debug(`_createOrUpdateProgram: ${oldTsProgram ? 'update' : 'create'} TypeScript Program`);
        this._program = this._ts.createProgram(this._rootNames, this._compilerOptions, this._compilerHost, oldTsProgram);
    }
    _transpileModule(fileContent, transpileOptions, customTransformers) {
        const diagnostics = [];
        const options = transpileOptions.compilerOptions
            ?
                this._ts.fixupCompilerOptions(transpileOptions.compilerOptions, diagnostics)
            : {};
        const defaultOptions = this._ts.getDefaultCompilerOptions();
        for (const key in defaultOptions) {
            if (this._ts.hasProperty(defaultOptions, key) && options[key] === undefined) {
                options[key] = defaultOptions[key];
            }
        }
        for (const option of this._ts.transpileOptionValueCompilerOptions) {
            options[option.name] = option.transpileOptionValue;
        }
        options.suppressOutputPathCheck = true;
        options.allowNonTsExtensions = true;
        const inputFileName = transpileOptions.fileName ||
            (transpileOptions.compilerOptions && transpileOptions.compilerOptions.jsx ? 'module.tsx' : 'module.ts');
        const sourceFile = this._ts.createSourceFile(inputFileName, fileContent, options.target);
        if (transpileOptions.moduleName) {
            sourceFile.moduleName = transpileOptions.moduleName;
        }
        if (transpileOptions.renamedDependencies) {
            sourceFile.renamedDependencies = new Map(getEntries(transpileOptions.renamedDependencies));
        }
        const newLine = this._ts.getNewLineCharacter(options);
        let outputText;
        let sourceMapText;
        const compilerHost = {
            getSourceFile: (fileName) => (fileName === this._ts.normalizePath(inputFileName) ? sourceFile : undefined),
            writeFile: (name, text) => {
                if (this._ts.fileExtensionIs(name, '.map')) {
                    this._ts.Debug.assertEqual(sourceMapText, undefined, 'Unexpected multiple source map outputs, file:', name);
                    sourceMapText = text;
                }
                else {
                    this._ts.Debug.assertEqual(outputText, undefined, 'Unexpected multiple outputs, file:', name);
                    outputText = text;
                }
            },
            getDefaultLibFileName: () => 'lib.d.ts',
            useCaseSensitiveFileNames: () => false,
            getCanonicalFileName: (fileName) => fileName,
            getCurrentDirectory: () => '',
            getNewLine: () => newLine,
            fileExists: (fileName) => fileName === inputFileName,
            readFile: () => '',
            directoryExists: () => true,
            getDirectories: () => [],
        };
        const program = this._ts.createProgram([inputFileName], options, compilerHost);
        if (this.ngJestConfig.shouldReportDiagnostics(inputFileName)) {
            this._ts.addRange(diagnostics, program.getSyntacticDiagnostics(sourceFile));
            this._ts.addRange(diagnostics, program.getOptionsDiagnostics());
        }
        program.emit(undefined, undefined, undefined, undefined, this.createTransformers(customTransformers, program));
        if (outputText === undefined)
            return this._ts.Debug.fail('Output generation failed');
        return { outputText, diagnostics, sourceMapText };
    }
    createTransformers(customTransformers, program) {
        return {
            before: customTransformers.before.map((beforeTransformer) => beforeTransformer.factory({ configSet: this.ngJestConfig, program }, beforeTransformer.options)),
            after: customTransformers.after.map((afterTransformer) => afterTransformer.factory({ configSet: this.ngJestConfig, program }, afterTransformer.options)),
            afterDeclarations: customTransformers.afterDeclarations.map((afterDeclarations) => afterDeclarations.factory({ configSet: this.ngJestConfig, program }, afterDeclarations.options)),
        };
    }
}
exports.NgJestCompiler = NgJestCompiler;
