"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NgJestTransformer = void 0;
const ts_jest_transformer_1 = require("ts-jest/dist/ts-jest-transformer");
const json_1 = require("ts-jest/dist/utils/json");
const jsonable_value_1 = require("ts-jest/dist/utils/jsonable-value");
const ng_jest_compiler_1 = require("./compiler/ng-jest-compiler");
const ng_jest_config_1 = require("./config/ng-jest-config");
class NgJestTransformer extends ts_jest_transformer_1.TsJestTransformer {
    process(fileContent, filePath, transformOptions) {
        return super.process(fileContent, filePath, transformOptions);
    }
    _configsFor(transformOptions) {
        const jestConfig = Object.assign({}, transformOptions.config);
        const ccs = NgJestTransformer._cachedConfigSets.find((cs) => cs.jestConfig.value === jestConfig);
        let ngJestConfig;
        if (ccs) {
            this._transformCfgStr = ccs.transformerCfgStr;
            this._compiler = ccs.ngJestCompiler;
            ngJestConfig = ccs.ngJestConfig;
        }
        else {
            const stringifiedJestCfg = json_1.stringify(jestConfig);
            const serializedCcs = NgJestTransformer._cachedConfigSets.find((cs) => cs.jestConfig.serialized === stringifiedJestCfg);
            if (serializedCcs) {
                serializedCcs.jestConfig.value = jestConfig;
                this._transformCfgStr = serializedCcs.transformerCfgStr;
                this._compiler = serializedCcs.ngJestCompiler;
                ngJestConfig = serializedCcs.ngJestConfig;
            }
            else {
                this._logger.info('no matching config-set found, creating a new one');
                ngJestConfig = new ng_jest_config_1.NgJestConfig(jestConfig);
                this._compiler = new ng_jest_compiler_1.NgJestCompiler(ngJestConfig, transformOptions.cacheFS);
                this._transformCfgStr = new jsonable_value_1.JsonableValue(Object.assign(Object.assign({}, jestConfig), ngJestConfig.parsedTsConfig)).serialized;
                NgJestTransformer._cachedConfigSets.push({
                    jestConfig: new jsonable_value_1.JsonableValue(jestConfig),
                    ngJestConfig,
                    transformerCfgStr: this._transformCfgStr,
                    ngJestCompiler: this._compiler,
                });
                this._getFsCachedResolvedModules(ngJestConfig);
            }
        }
        return ngJestConfig;
    }
}
exports.NgJestTransformer = NgJestTransformer;
NgJestTransformer._cachedConfigSets = [];
